""""
directory_paths.py

Define functions to get common hydrogen directory paths.
"""
import os
import json
import datetime
import logging

def get_hydrodata_directory():
    """Returns the full path of the hydroframe /hydrodata directory containing large files."""
    result = os.environ.get("HYDRODATA", None)
    if not result or not os.path.exists(result):
        # If the HYDRODATA environment variable is not used use default
        result = "/hydrodata"

    return result


def get_hydro_common_directory():
    """Returns the common directory in /home/HYDROAPP/common that contains common static files for all users."""

    # The 'standard' place is different depending upon whether code is running on the client, the VM host or a Docker container
    # It is one of the directories specified by these environment variables
    result = None
    hydrocommon = os.environ.get("HYDROCOMMON", "")
    directory_options = [hydrocommon, "/hydrocommon", "/home/HYDROAPP/common"]
    for dirpath in directory_options:
        if os.path.exists(dirpath):
            result = dirpath
            break
    return result


def get_data_directory():
    """Returns the full path name of the data directory where files are stored, or None if not configured."""

    # The 'standard' place is different depending upon whether code is running on the client, the VM host or a Docker container
    # It is one of the directories specified by these environment variables
    result = None
    directory_env_variables = [
        "CONTAINER_HYDRO_DATA_PATH",
        "CLIENT_HYDRO_DATA_PATH",
        "HOST_HYDRO_DATA_PATH",
    ]
    for env_var in directory_env_variables:
        dirpath = os.environ.get(env_var, None)
        if dirpath is not None and os.path.exists(dirpath):
            result = dirpath
            break
    return result


def get_domain_path(user_id=None, domain_id=None, domain_directory=None, message=None):
    """
    Returns the full path name to the domain directory for the environment where code is running.
    """

    if message is not None:
        user_id = message.get("user_id", None)
        domain_id = message.get("domain_id", None)
        domain_directory = message.get("domain_directory", None)
    if user_id is None:
        raise Exception("No user_id provided.")
    if domain_directory and domain_id is None:
        # For backward compatibility
        domain_id = domain_directory
    if domain_id is None:
        raise Exception("No domain_id provided.")
    domain_id = domain_id.lower()
    user_id = user_id.lower()
    data_dir = get_data_directory()
    domain_path = f"{data_dir}/{user_id}/{domain_id}"
    return domain_path


def get_domain_state(user_id=None, domain_id=None, domain_directory=None, message=None):
    """
    Return the contents of the domain_state.json object of the domain directory.
    """

    result = None
    if message is not None:
        user_id = message.get("user_id", None)
        domain_id = message.get("domain_id", None)
        domain_directory = message.get("domain_directory", None)
    if user_id is None:
        raise Exception("No user_id provided.")
    if domain_directory and domain_id is None:
        # For backward compatibility
        domain_id = domain_directory
    if domain_id is None:
        raise Exception("No domain_id provided.")

    domain_path = get_domain_path(user_id=user_id, domain_id=domain_id)
    domain_state_filename = f"{domain_path}/domain_state.json"
    if os.path.exists(domain_state_filename):
        with open(domain_state_filename, "r") as stream:
            domain_state = stream.read()
            result = json.loads(domain_state)
    # In case the directory was copied or moved put the actual user and directory in the returned state
    if result:
        result["user_id"] = user_id.lower()
        result["domain_id"] = domain_id.lower()
    return result

def update_domain_state(domain_state):
    """
        Update the domain state in the user domain directory with the attributes in the domain_state object.
        The domain_state object must contain at least the 'user_id' and 'domain_id' attributes.
        Other attributes in the object are replaced in the domain state in the user domain directory.
        Attributes not provided in the domain_state object are unchanged in the user domain directory.
    """

    lock_file_name = ""
    try:
        user_id = domain_state.get("user_id", None)
        domain_directory = domain_state.get("domain_directory", None)
        domain_id = domain_state.get("domain_id", None)
        if domain_directory and domain_id is None:
            # For backward compatibility
            domain_id = domain_directory

        if not user_id:
            raise Exception("No user_id in the domain state")
        if not domain_id:
            raise Exception("No domain_id in the domain state")
        domain_path = get_domain_path(user_id=user_id, domain_id=domain_id)
        domain_state_filename = f"{domain_path}/domain_state.json"
        lock_file_name = f"{domain_state_filename}.lck"
        prev_domain_state = {}

        # Create a lock file to protect against two processes writing at the same time
        with open(lock_file_name, "w+") as lock_stream:
            # Lock the lock file before we write the domain_state file
            if not os.name == "nt":
                # fcntl only works on linux so no locking on windows...
                try:
                    import fcntl

                    fcntl.flock(lock_stream.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                except Exception as e:
                    raise Exception("Domain state was locked. Unable to save") from e

            if os.path.exists(domain_state_filename):
                # Read the previous contents of the prev_domain_state file first
                with open(domain_state_filename, "r") as read_stream:
                    contents = read_stream.read()
                    if contents:
                        prev_domain_state = json.loads(contents)
            with open(domain_state_filename, "w+") as write_stream:
                # Update the attributes from the incoming message, leave the rest alone
                if (
                    prev_domain_state.get("state", None) == "deleted"
                    and prev_domain_state.get("state_before_deleted", None)
                    and domain_state.get("state", None) == "new"
                ):
                    # When un-deleting a domain restore the previous state
                    domain_state["state"] = prev_domain_state.get("state_before_deleted", None)

                # Update values in the new domain_state given the values from in input param
                for attribute_name in domain_state.keys():
                    if not attribute_name in ["user_id", "domain_id", "domain_directory"]:
                        new_value = domain_state.get(attribute_name, None)
                        prev_domain_state[attribute_name] = new_value

                if not prev_domain_state.get("state", None):
                    # If there is no state for the domain then set the state to 'new'
                    prev_domain_state["state"] = "new"
                if not prev_domain_state.get("created_date", None):
                    prev_domain_state["created_date"] = datetime.datetime.utcnow().strftime(
                        "%Y-%m-%dT%H:%M:%S.000Z"
                    )

                # Put the actual user_id and domain_id the prev_domain_state
                prev_domain_state["directory_name"] = domain_id
                prev_domain_state["domain_directory"] = domain_id
                prev_domain_state["user_id"] = user_id
                
                # Save the updated file
                write_stream.write(json.dumps(prev_domain_state, indent=2))
        if os.path.exists(lock_file_name):
            os.remove(lock_file_name)
        return prev_domain_state
    except Exception as e:
        logging.exception("Unable to save domain state")
        if os.path.exists(lock_file_name):
            os.remove(lock_file_name)
        raise Exception("Unable to save domain state") from e


def get_domain_database(message):
    """
    Deprecated!
    Return the contents of the databse.json object of the domain directory.
    Use the user_id and domain_directory values in the message dict.
    This is deprecated. use get_domain_state() instead.
    """

    return get_domain_state(message)


def get_widget(message):
    """
    Deprecated!
    Return the visualization widget from the domain database associated with the
    domain and widget_id defined by the user_id, domain_directory and widget_id from
    the message dict.
    """

    result = None
    widget_id = message.get("widget_id", None)
    database = get_domain_database(message)
    if database is not None:
        for vis_index, vis in enumerate(database.get("visualizations", [])):
            for w_index, w in enumerate(vis.get("widgets", [])):
                if f"{vis_index}.{w_index}" == widget_id:
                    result = w
                    break
            if result:
                break
    return result



"""
job_main
    Defines a function to be used in files that implement code
    to handle message bus events. This function can be used
    to wait for message bus event either by subscribing to the event
    and waiting or by just waiting for the process to be invoked.

    This method looks for the -listen command line argument to know
    which method to use.
"""
import sys
import logging
from .use_message_bus import listen_on_message_bus
from .run_job_using_callback import run_job_using_callback


def job_main(subject, callback, env_variables=None):
    """Wait for events either by listening on the message bus or waiting for Argo to start process"""

    if sys.argv and len(sys.argv) > 1 and sys.argv[1] == '-listen':
        logging.info("Starting service listening on NATS for %s.", subject)
        listen_on_message_bus(subject, callback)
    else:
        env_variables = env_variables if env_variables else []
        logging.info("Running %s workflow for Argo.", subject)
        run_job_using_callback(subject, callback, env_variables)



"""
use_message_bus

    Define functions to use the message bus.
    These functions simplify the usage of the message bus.
"""
import os
import asyncio
import logging
import json
import inspect
from .message_bus import MessageBus

# pylint: disable=C0103,W0603
# Variable to allow dependency injection of MessageBus implementation
MessageBusClass = MessageBus


def listen_on_message_bus(subject, callback):
    """Subscribe and listen forever for events with with a subject.

    The callback must be a async function that is called with the message
    as an argument when an incoming message is received.
    """

    try:
        asyncio.run(__event_loop__(subject, callback))
    except Exception:
        pass


async def __event_loop__(subject, callback):
    """Add the subscription to the event bus using the MessageBus class."""

    async with MessageBusClass() as bus:
        sid = None
        try:
            sid = await bus.subscribe(subject, cb=callback)
            logging.info("Subscribed to '%s' events.", subject)
            await bus.wait()
        except Exception:
            logging.exception("Exit wait loop for '%s' because of exception.", subject)
            if sid is not None:
                logging.info("Unsubscribed to subject '%s'.", subject)
                await bus.unsubscribe(sid)


async def execute_with_class_and_reply(class_instance, subject, reply, json_message):
    """
    Execute a method from the class_instance with the json_message as an argument.
    Use the value of the 'op' attribute of the message to find the method name to execute.
    If there is no method with the value of the 'op' attribute execute the method handle_event().

    Publish a reply when finished with the value method result or an error message.
    The subject is specified only to log the subject for which the reply is sent.
    If reply is None, then no reply is published after the completion of the method call.
    """

    try:
        logging.debug("Received message subject '%s' %s", subject, json_message)
        container_hydro_data_path = os.environ.get("CONTAINER_HYDRO_DATA_PATH", None)
        client_hydro_data_path = os.environ.get("CLIENT_HYDRO_DATA_PATH", None)
        if container_hydro_data_path is None or (
            not os.path.exists(container_hydro_data_path)
            and os.path.exists(client_hydro_data_path)
        ):
            # This means the job is probably running locally and not within a container
            # So set the CONTAINER_HYDRO_DATA_PATH so the job will read/write to the client_hydro_data_path
            os.environ["CONTAINER_HYDRO_DATA_PATH"] = client_hydro_data_path
        op = json_message.get('op', None)
        if op is not None and hasattr(class_instance, op):
            # Invoke the method with the value of the 'op' attribute
            f = getattr(class_instance, op)
            if inspect.iscoroutinefunction(f):
                response = await f(json_message)
            else:
                response = f(json_message)
        elif hasattr(class_instance, 'handle_event'):
            # Otherwise invoke the handle_event method
            if inspect.iscoroutinefunction(class_instance.handle_event):
                response = await class_instance.handle_event(json_message)
            else:
                response = class_instance.handle_event(json_message)
        elif op:
            raise Exception(
                f"No method '{op}' defined in class '{type(class_instance).__name__}'."
            )
        else:
            raise Exception(
                f"No method handle_event defined in class '{type(class_instance).__name__}'."
            )
        if reply:
            # Publish a message with reply as the subject with the response
            response = {} if response is None else response
            if isinstance(response, dict):
                response['status'] = 'success'
                response['request'] = json_message
                response['subject'] = subject
            logging.debug(
                "Reply back with subject '%s' from message subject '%s'.",
                reply, subject
            )
            async with MessageBusClass() as bus:
                await bus.publish(reply, json.dumps(response))
    except Exception as e:
        logging.exception("Exception ocurred while handling message %s.", subject)
        async with MessageBusClass() as bus:
            await bus.publish(reply, json.dumps({'status': 'fail', 'message': str(e), 'subject': subject, 'request': json_message}))


async def execute_with_function_and_reply(callback, subject, reply, json_message):
    """Execute the callback function with the json_message as an argument
    Publish a the response with the reply subject with the function result or an error message.
    """

    try:
        logging.debug("Received message subject '%s' %s", subject, json_message)
        container_hydro_data_path = os.environ.get("CONTAINER_HYDRO_DATA_PATH", None)
        client_hydro_data_path = os.environ.get("CLIENT_HYDRO_DATA_PATH", None)
        if not os.path.exists(container_hydro_data_path) and os.path.exists(
            client_hydro_data_path
        ):
            # This means the job is probably running locally and not within a container
            # So set the CONTAINER_HYDRO_DATA_PATH so the job will read/write to the client_hydro_data_path
            os.environ["CONTAINER_HYDRO_DATA_PATH"] = client_hydro_data_path

        if inspect.iscoroutinefunction(callback):
            # this is an async function so call it with await
            response = await callback(json_message)
        else:
            response = callback(json_message)

        if reply:
            response = {} if response is None else response
            if isinstance(response, dict):
                response['status'] = 'success'
                response['request'] = json_message
                response['subject'] = subject
            async with MessageBusClass() as bus:
                await bus.publish(reply, json.dumps(response))
    except Exception as e:
        logging.exception("Exception ocurred while handling message %s.", subject)
        async with MessageBusClass() as bus:
            await bus.publish(reply, json.dumps({'status': 'fail', 'message': str(e), 'subject': subject, 'request': json_message}))


async def send_request_message(message, subject):
    """Publish a synchronous request message and return the replied result."""

    response = None
    try:
        async with MessageBusClass() as bus:
            reply = await bus.request(subject, json.dumps(message))
            response = json.loads(reply)
    except Exception:
        error_message = f"Exception while sending request '{subject}' message to NATS"
        logging.exception(error_message)
        response = {'status': 'fail', 'message': error_message}
    return response


async def send_publish_message(message, subject):
    """Publish an asynchronous message with the subject to message bus."""

    response = None
    try:
        async with MessageBusClass() as bus:
            await bus.publish(subject, json.dumps(message))
            port = os.environ.get("NATS_PORT", None)
            server = os.environ.get("NATS_SERVER", None)
            response = {'message': f"sent '{subject}' subject to NATS {server}:{port}"}
    except Exception:
        error_message = f"Exception while sending '{subject}' message to NATS"
        logging.exception(error_message)
        response = {'status': 'fail', 'message': error_message}
    return response


def set_message_bus_class(c):
    """Set the message bus class to allow dependency injection for unit testing."""
    global MessageBusClass
    MessageBusClass = c



"""
message_bus

    Defines a class to provide operations to use the message bus used by hydrogen.
    This is an abstraction to message bus that is implemented to use NAT, but the same
    abstrction could be used to implement with other message bus implementation.
"""
import os
import asyncio
import logging
import socket
import nats

NATS_PORT = os.getenv("NATS_PORT", "4222")
NATS_SERVER = os.getenv("NATS_SERVER", "localhost")


class MessageBus:
    """
    Service for reading and writing to the message bus.
    This insulates the caller from the choice of NATS as the message bus implementation
    and simplies the interface to what is used in the hydrogen project.

    Use the class as a context. For example, to publish a message on the bus.
        async with MessageBus() as bus:
            await bus.publish('message', 'body')

    This code must be within an async function that is called directly or indirectly from a
        asyncio.run(...method_or_function...)
    The above call can be made from anywhere such as a flask resource handler.
    """

    def __init__(self):
        """Constructor."""
        self.nc = None
        self.never = None

    async def __aenter__(self):
        """Python context asyncronous enter event. Create a NATS connection."""

        url = f"nats://{NATS_SERVER}:{NATS_PORT}"
        logging.debug(
            "Create MessageBus instance and connect to NATS server with with url '%s'.",
            url
        )
        self.nc = await nats.connect(url)
        self.never = asyncio.Future()
        return self

    async def __aexit__(self, *args, **kwargs):
        """
        Python context asyncronous exit event.
        Wait for all NATS messages to process before exit.
        """

        await self.nc.flush()
        await self.nc.drain()

    async def publish(self, subject, message):
        """public an asyncronous message to the bus. The subject and message are strings."""

        logging.debug("Publish subject '%s' %s", subject, message)
        await self.nc.publish(subject, message.encode())

    async def request(self, subject, message, timeout=10):
        """Publish a synchronous message to the bus and return the reply."""

        logging.debug("Request subject '%s' %s", subject, message)
        response = await self.nc.request(subject, message.encode(), timeout=timeout)
        return response.data.decode()

    async def subscribe(self, subject, cb):
        """
        Subscribe to receive message for a subject using a cb (callback method).
        The subject is also used as a queue so only one subscriber handles a message to allow for
        load balancing of messages.

        The callback is a callback function (it must be an async function, not a method in a class).
        """
        logging.info("Subscribe to subject '%s'.", subject)
        sid = await self.nc.subscribe(subject, cb=cb, queue=subject)
        return sid

    async def wait(self):
        """Wait forever while still listening for subscriptions."""

        hostname = socket.gethostname()
        logging.info("Wait for messages for '%s'.", hostname)
        await (self.never)



"""
run_job_using_callback

    Define a function to run a job using an async callback function.
    This is intended to support asynchronous jobs such as Argo
    workflow jobs.
"""

import os
import asyncio
import logging
import json


def run_job_using_callback(subject, callback, environment_variables):
    """Run a job using an asynchronous callback methods.
    Pass specified environment variable values as arguments.
    """

    try:
        # Create a NATS message with the environment variable values as data
        class Message:
            """Abstraction for a message used in call to callback."""
            def __init__(self):
                self.subject = None
                self.data = None

        message = Message()
        message.subject = subject
        data = {}
        for var in environment_variables:
            data[var] = os.environ.get(var, None)
        message.data = json.dumps(data).encode()

        # Set the CONTAINER_HYDRO_DATA_PATH env variable to work in DOCKER and locally
        container_hydro_data_path = os.environ.get("CONTAINER_HYDRO_DATA_PATH", None)
        client_hydro_data_path = os.environ.get("CLIENT_HYDRO_DATA_PATH", None)
        if not os.path.exists(container_hydro_data_path) and os.path.exists(
            client_hydro_data_path
        ):
            # This means the job is probably running locally and not within a container
            # So set the CONTAINER_HYDRO_DATA_PATH so the job will read/write to the client_hydro_data_path
            os.environ["CONTAINER_HYDRO_DATA_PATH"] = client_hydro_data_path

        # Invoke the NATS callback method to run the job and send back the reply
        asyncio.run(callback(message))
    except Exception:
        logging.exception(
            "Unable to run_job_subscribe_callback with subject '%s'.",
            subject,
        )



"""
run_job.py

    Define functions to start an asynchonous job and to execute a job
    through an SSH tunnel from a client such as the API.
"""
import os
import asyncio
import logging
import json
import requests

from .directory_paths import update_domain_state
from .use_message_bus import send_publish_message


def start_async_job(job_name, user_id, domain_id, parameters=None):
    """Start an asyncrounous job to run in background."""

    job_server = os.environ.get("JOB_SERVER", None)
    job_server_port = os.environ.get("JOB_SERVER_PORT", None)
    nats_port = os.environ.get("NATS_PORT", None)

    if job_server and job_server_port:
        # An ssh tunnel is configured. So use it to execute remote jobs
        logging.info(
            "Sent http request '%s' for '%s' of user '%s'.",
            job_name,
            domain_id,
            user_id,
        )
        # the ssh tunnel executes the job in a background process so returns fast
        run_remote_job(job_name, user_id, domain_id, parameters)
    elif nats_port:
        # A NATS port is configured to send a NATS message to Argo to execute remote job
        try:
            message = {"user_id": user_id, "domain_id": domain_id, "domain_directory": domain_id}
            if parameters:
                for key in parameters.keys():
                    value = parameters.get(key, None)
                    message[key] = value
            asyncio.run(send_publish_message(message, job_name))
            logging.info(
                "Published '%s' to NATS bus using '%s' for user '%s'",
                job_name,
                domain_id,
                user_id,
            )
            logging.info(message)
        except Exception:
            logging.exception("Error when trying to run remote Argo job")
            error_message = f"Unable to run job '{job_name}"
            domain_state = {
                "user_id": user_id,
                "domain_id": domain_id,
                "state": "error",
                "error_message": error_message,
            }
            update_domain_state(domain_state)


def run_remote_job(job_name, user_id, domain_id, parameters=None):
    """Start an job on a remote server using an ssh tunnel."""

    job_server = os.environ.get("JOB_SERVER", None)
    job_server_port = os.environ.get("JOB_SERVER_PORT", None)
    if job_server and job_server_port:
        url = f"http://{job_server}:{job_server_port}/run_job/{user_id}/{domain_id}/{job_name}"
        if parameters:
            parameter_values = [
                f"{key}={parameters.get(key)}" for key in parameters.keys()
            ]
            url = url + "?" + "&".join(parameter_values)
        logging.info(url)
        response = requests.get(url)
        if not response.status_code == 200:
            message = response.text
            raise Exception("Job {job_name} failed with error: {message}")
        if response.text:
            response_json = json.loads(response.text)
            if response_json.get("status", None) == "fail":
                message = response_json.get("message", None)
                raise Exception(f"Job {job_name} failed with error: {message}")
    else:
        raise Exception("No configuration for executing remote job.")
